%
% $Id: varda_controller.pl,v 1.1 2009-08-14 11:20:19 gjn Exp $
%
%     Copyright 2009 by Pawel Plazienski 2009
%
%     VARDA shell predicates
%
%
%     Copyright (C) 2006-9 by the HeKatE Project
%
%     VARDA has been develped by the HeKatE Project,
%     see http://hekate.ia.agh.edu.pl
%
%     This file is part of VARDA.
%
%     VARDA is free software: you can redistribute it and/or modify
%     it under the terms of the GNU General Public License as published by
%     the Free Software Foundation, either version 3 of the License, or
%     (at your option) any later version.
%
%     VARDA is distributed in the hope that it will be useful,
%     but WITHOUT ANY WARRANTY; without even the implied warranty of
%     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%     GNU General Public License for more details.
%
%     You should have received a copy of the GNU General Public License
%     along with VARDA.  If not, see <http://www.gnu.org/licenses/>.


:- write('VARDA: $Id: varda_controller.pl,v 1.1 2009-08-14 11:20:19 gjn Exp $'),nl.

:- ensure_loaded('varda_read_tools').

shl :-
	get_line(Line,
		[sar,shi,sha,hic,tic,gax,sxt,pxt,xop,kid,axg,xxg,arp,pur,pux,hlp,hlt,ada,adp,dep,fin,spl]),
	shell_interpret_line(Line).

shl.

shell_interpret_line(Line) :-
	Line = end_of_file.

shell_interpret_line(Line) :-
	empty_line(Line),
	!, shl.

shell_interpret_line(Line) :-
	string_to_atom(Line,Name),
	current_functor(Name,0),
	(call(Name) ->
		write('<+- : Success'), nl;
		write('<!- : Failure'), nl
	), !, shl.

shell_interpret_line(Line) :-
	string_to_atom(Line,Name),
	not(current_functor(Name,0)),
	write('<!- : No such function'), nl,
	!, shl.

ada :-
	write('Attribute name'),
	get_line(Text),
	atomize(Text,Atom),
	ard_att_add(Atom).

adp :-
	write('Property attribute'),
	ask_attribute(Attribute),
	ard_property_add([Attribute]).

dep :-
	write('From (full property)'),
	ask_property(From),
 	write('To (full property)'),
	ask_property(To),
	ard_depend_add(From,To).

fin :-
	write('Old (simple property)'),
	ask_simple(OldProperty),
	write('New (attributes)'),
	ask_attribute_list(NewProperty),
	ard_finalize(OldProperty,NewProperty).

spl :-
	write('Old (full property)'),
	ask_property(OldProperty),
	write('New (full property list)'),
	ask_list(NewPropertyList),
	ard_split(OldProperty,NewPropertyList,[]).

ask_attribute(Attribute) :-
	get_attributes(Atts),
	get_line(Text,Atts),
	atomize(Text,Attribute).

ask_attributes(Attributes) :-
	get_attributes(Atts),
	get_line(Text,Atts),
	atomize(Text,Attributes).

ask_simple(Property) :-
	get_simple_properties(Props),
	properties_to_strings(Props,PropStrings),
	get_line(Text,PropStrings),
	not(empty_line(Text)),
	atomize(Text,Atom),
	propertize([Atom],Property).

ask_property(Property) :-
	get_properties(Props),
	properties_to_strings(Props,PropStrings),
	get_line(Text,PropStrings),
	not(empty_line(Text)),
	string_list(Text,List,","),
	atomize(List,Atoms),
	propertize(Atoms,Property).

ask_list(List) :-
	ask_list(List,[]).

ask_list(List,Got) :-
	get_properties(Props),
	properties_to_strings(Props,PropStrings),
 	get_line(Text,PropStrings),
 	(  empty_line(Text),
		List = Got
	;  not(empty_line(Text)),
		string_list(Text,Strings,","),
		atomize(Strings,Current),
		propertize(Current,Property),
		union(Got,[Property],G),
		ask_list(List,G)
	).

ask_attribute_list(List) :-
	ask_attribute_list(List,[]).

ask_attribute_list(List,Got) :-
	get_attributes(Atts),
 	get_line(Text,Atts),
 	(  empty_line(Text),
		List = Got
	;  not(empty_line(Text)),
		atomize(Text,Atom),
		union(Got,[Atom],G),
		ask_attribute_list(List,G)
	).


simple_property(X) :-
	X = [_].

get_attributes(List) :-
	setof(X, ard_att(X), List).

get_properties(List) :-
	setof(X, ard_property(X), List).

get_properties([]).

get_simple_properties(List) :-
	get_properties(All),
	setof(X, (member(X,All), simple_property(X)), List), !.

get_simple_properties([]).

properties_to_strings([HProp|Props],[HString|Strings]) :-
	property_to_string(HProp,HString),
	properties_to_strings(Props,Strings).

properties_to_strings([],[]).

property_to_string(Property,String) :-
	list_string(Property,String,',').

propertize(List,Prop) :-
	sort(List,Prop).

atomize(Text,Atom) :-
	\+ is_list(Text),
	string_to_list(Text,X),
	atom_chars(Atom,X).

atomize([HText|Texts],[HAtom|Atoms]) :-
	atomize(Texts,Atoms),
	atomize(HText,HAtom).

atomize([],[]).
