%
% $Id: varda_read_tools.pl,v 1.1 2009-08-14 11:20:19 gjn Exp $
%
%     Copyright 2009 by Pawel Plazienski 2009
%
%     VARDA read tools
%
%
%     Copyright (C) 2006-9 by the HeKatE Project
%
%     VARDA has been develped by the HeKatE Project,
%     see http://hekate.ia.agh.edu.pl
%
%     This file is part of VARDA.
%
%     VARDA is free software: you can redistribute it and/or modify
%     it under the terms of the GNU General Public License as published by
%     the Free Software Foundation, either version 3 of the License, or
%     (at your option) any later version.
%
%     VARDA is distributed in the hope that it will be useful,
%     but WITHOUT ANY WARRANTY; without even the implied warranty of
%     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%     GNU General Public License for more details.
%
%     You should have received a copy of the GNU General Public License
%     along with VARDA.  If not, see <http://www.gnu.org/licenses/>.


:- write('VARDA: $Id: varda_read_tools.pl,v 1.1 2009-08-14 11:20:19 gjn Exp $'),nl.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
list_string([],S,_) :-
	S = ''.

list_string(List,String,Separator) :-
	List = [Head|Tail],
	list_string(Tail,Rest,Separator),
	string_to_atom(Element,Head),
	(  Tail = [],
		Rest = PrepRest
	;  not(Tail = []),
		string_concat(Separator,Rest,PrepRest)
	),
	string_concat(Element,PrepRest,String).

lists_strings([],[]).

lists_strings(ListifiedList,StringList) :-
	ListifiedList = [ListifiedHead|ListifiedTail],
	StringList = [StringHead|StringTail],
	string_to_list(StringHead,ListifiedHead),
	lists_strings(ListifiedTail,StringTail).

string_list(String,List,[Separator]) :-
	string_to_list(String,Listified),
	list_separated(Listified,ListifiedLists,Separator),
	lists_strings(ListifiedLists,List).

list_separated(Separated,Lists,Separator) :-
	not(member(Separator,Separated)),
	Lists = [Separated].

list_separated(Separated,Lists,Separator) :-
	append(First,[Separator|Rest],Separated),
	not(member(Separator,First)),
	list_separated(Rest,Tail,Separator),
	Lists = [First|Tail].

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

get_line(Text) :-
	get_line(Text,[]).

get_line(Text,Completions) :-
	get_line(Text,Completions,'-.> ').

get_line(Text,Completions,Prompt) :-
	write(Prompt),
	get_line([],ReverseString,Completions,Prompt),
	reverse(ReverseString,String),
	string_to_list(Text,String).

get_line(Got,ReverseString,Completions,Prompt) :-
	get_single_char(B),
% 	write(B),
	( (B = 13 ; B = 10),
		nl, ttyflush,
		Got = ReverseString, !
	; B = 9,
		reverse(Got,ReverseGot),
		string_to_atom(ReverseGot,Begin),
		find_begin(Completions,Begin,Begins),
		( Begins = [],
			nl, write('<!- No completions'),
			nl, write(Prompt),
			(  Begin = []
			;  not(Begin = []),
			write(Begin)
			),
			get_line(Got,ReverseString,Completions,Prompt)
		; Begins = [Single],
			nl, write(Prompt), write(Single),
			string_to_atom(S,Single),
			string_to_list(S,G),
			reverse(G,ReverseG),
			get_line(ReverseG,ReverseString,Completions,Prompt)
		; Begins = [_|Tail], not(Tail = []),
			nl, pretty_write_list(Begins), write(Prompt),
			(  Begin = []
			;  not(Begin = []),
			write(Begin)
			),
			get_line(Got,ReverseString,Completions,Prompt)
		)
	; B = 127,
		(  Got = [_|G],
		put_byte(8),
		put_byte(32),
		put_byte(8),
		get_line(G,ReverseString,Completions,Prompt)
		;  Got = [],
		get_line(Got,ReverseString,Completions,Prompt)
		)
	; B = -1,
		nl, ttyflush,
		Got = end_of_file, !
	; not(B = 13 ; B = 10 ; B = 127; B = 9; B = -1),
		char_code(C,B),
		put_char(C),
		G = [B|Got],
		get_line(G,ReverseString,Completions,Prompt)
	).

empty_line(X) :-
	string_to_atom(X,'').

find_begin(Completions,[],Completions).

find_begin(Completions,Begin,Begins) :-
	bagof(X, (member(X,Completions), begins(Begin,X)), Begins), !.

find_begin(_,_,[]).

begins(Begin,String) :-
	string_concat(Begin,_,String).

pretty_write_list([]).
pretty_write_list([H|T]) :-
	write('<.- : '), write(H), nl,
	pretty_write_list(T).
